from dataclasses import dataclass

from pants.engine.env_vars import EnvironmentVars as EnvironmentVars
from pants.engine.env_vars import EnvironmentVarsRequest as EnvironmentVarsRequest
from pants.engine.rules import Get, collect_rules, rule
from pants.engine.target import FieldSet
from pants.engine.unions import UnionRule

from pants_backend_secrets.exception import MissingSecret
from pants_backend_secrets.goals.decrypt import DecryptFieldSet, DecryptRequest, DecryptResponse
from pants_backend_secrets.secret_request import (
    FallibleSecretsRequest,
    FallibleSecretsResponse,
    SecretsResponse,
    SecretValue,
)
from pants_backend_secrets.targets import EnvironmentSecretKey


@dataclass(frozen=True)
class EnvironmentSecretResponse(SecretsResponse):
    def cacheable(self) -> bool:
        return False


@dataclass(frozen=True)
class EnvironmentSecretFieldSet(FieldSet):
    required_fields = (EnvironmentSecretKey,)
    key: EnvironmentSecretKey


@dataclass(frozen=True)
class FallibleEnvironmentSecretsRequest(FallibleSecretsRequest):
    field_set_type = EnvironmentSecretFieldSet


@rule
async def get_environment_key(
    request: FallibleEnvironmentSecretsRequest,
) -> FallibleSecretsResponse:
    relevant_env = await Get(EnvironmentVars, EnvironmentVarsRequest([request.target.key.value]))

    if request.target.key.value not in relevant_env:
        return FallibleSecretsResponse(
            exit_code=1,
            stdout=f"secret is not in host environment: `{request.target.key.value}`",
        )

    # no stdout or stderr on exit 0, to avoid leaks
    return FallibleSecretsResponse(
        exit_code=0,
        response=SecretsResponse(
            SecretValue(relevant_env[request.target.key.value], request.target.address, "EnvironmentVars"),
        ),
    )


class DecryptEnvironmentVarsRequest(DecryptRequest):
    pass


@dataclass(frozen=True)
class DecryptEnvironmentFieldSet(DecryptFieldSet):
    decrypt_request_type = DecryptEnvironmentVarsRequest
    required_fields = (EnvironmentSecretKey,)

    key: EnvironmentSecretKey


@rule
async def decrypt_environment(request: DecryptEnvironmentVarsRequest) -> DecryptResponse:
    response = await Get(FallibleSecretsResponse, FallibleEnvironmentSecretsRequest(request.field_set))

    if response.exit_code != 0:
        raise MissingSecret(f"failed retrieving environment secret: {response.stdout}")

    return DecryptResponse(response.response.value)


def rules():
    return [
        *collect_rules(),
        UnionRule(
            FallibleSecretsRequest,
            FallibleEnvironmentSecretsRequest,
        ),
        *DecryptEnvironmentFieldSet.rules(),
    ]



from __future__ import annotations

from dataclasses import dataclass
from typing import ClassVar

from pants.engine.engine_aware import EngineAwareReturnType
from pants.engine.environment import EnvironmentName
from pants.engine.rules import collect_rules, rule
from pants.engine.target import FieldSet, Target
from pants.engine.unions import UnionMembership, union
from pants.util.strutil import bullet_list

from pants_backend_secrets.exception import FailedDecryptException


@dataclass(frozen=True)
class SecretValue:
    """Wrapped for a secret which redacts its value when printed"""

    value: str
    """The value of the secret"""

    name: str | None = None
    """The name of the secret. Used in lieu of the secret value in logs."""

    source: str | None = None
    """The source of the secret - such as the tool or system owning it. Only used in logs."""

    def __repr__(self):
        return f"Secret(<redacted>, {repr(self.name)}, {repr(self.source)})"

    def __str__(self):
        msg = "Secret"
        if self.name:
            msg += f" `{self.name}`"

        if self.source:
            msg += f" from `{self.source}`"

        return f"<{msg}>"


@union(in_scope_types=[EnvironmentName])
@dataclass(frozen=True)
class FallibleSecretsRequest:
    target: Target

    field_set_type: ClassVar[FieldSet]


@dataclass(frozen=True)
class SecretsRequestRequest:
    target: Target


@dataclass(frozen=True)
class SecretsRequestWrap:
    request: FallibleSecretsRequest | None


@dataclass(frozen=True)
class SecretsResponse:
    value: SecretValue

    def __post_init__(self):
        assert isinstance(self.value, SecretValue), f"value must be SecretValue but was {type(self.value)}"

    def cacheable(self) -> bool:
        return False


@dataclass(frozen=True)
class FallibleSecretsResponse(EngineAwareReturnType):
    exit_code: int
    stdout: str | None = None
    stderr: str | None = None
    response: SecretsResponse | None = None

    def cacheable(self) -> bool:
        if self.exit_code != 0:
            return True

        return self.response.cacheable()


@rule
def secrets_request_request(
    request: SecretsRequestRequest, union_membership: UnionMembership
) -> SecretsRequestWrap:
    tgt = request.target

    concrete_requests = [
        request_type(request_type.field_set_type.create(tgt))
        for request_type in union_membership[FallibleSecretsRequest]
        if request_type.field_set_type.is_applicable(tgt)
    ]

    if len(concrete_requests) != 1:
        raise ValueError(
            f"Multiple or zero registered decrypters from {SecretsRequestRequest.__name__} can "
            f"build target {tgt.name}. It is ambiguous which implementation to use.\n\n"
            "Possible implementations:\n\n"
            f"{bullet_list(sorted(generator.__name__ for generator in concrete_requests))}"
        )

    if len(concrete_requests) == 0:
        return SecretsRequestWrap(None)

    first_concrete = concrete_requests[0]
    return SecretsRequestWrap(first_concrete)


@rule
def unwrap_fallible_secret(fallible: FallibleSecretsResponse) -> SecretsResponse:
    if fallible.exit_code != 0:
        raise FailedDecryptException(
            "Failed decrypting secret",
            fallible.stdout,
            fallible.stderr,
        )

    return fallible.response


def rules():
    return [
        *collect_rules(),
    ]



""" """

from __future__ import annotations

from abc import ABCMeta
from dataclasses import dataclass
from typing import ClassVar, Generic, Type, TypeVar

from pants.engine.console import Console
from pants.engine.environment import EnvironmentName
from pants.engine.goal import Goal, GoalSubsystem, Outputting
from pants.engine.rules import Get, MultiGet, collect_rules, goal_rule
from pants.engine.target import (
    FieldSet,
    ImmutableValue,
    NoApplicableTargetsBehavior,
    TargetRootsToFieldSets,
    TargetRootsToFieldSetsRequest,
    Targets,
)
from pants.engine.unions import UnionMembership, UnionRule, union
from pants.util.frozendict import FrozenDict

from pants_backend_secrets.secret_request import SecretValue


class DecryptSubsystem(Outputting, GoalSubsystem):
    name = "decrypt"
    help = "A decrypt goal for secrets."


class Decrypt(Goal):
    subsystem_cls = DecryptSubsystem

    environment_behavior = Goal.EnvironmentBehavior.LOCAL_ONLY


_F = TypeVar("_F", bound=FieldSet)


@union(in_scope_types=[EnvironmentName])
@dataclass(frozen=True)
class DecryptRequest(Generic[_F]):
    field_set: _F


@dataclass(frozen=True)
class DecryptResponse:
    secret: SecretValue


_T = TypeVar("_T", bound=DecryptRequest)


class DecryptOutputData(FrozenDict[str, ImmutableValue]):
    pass


@union(in_scope_types=[EnvironmentName])
@dataclass(frozen=True)
class DecryptFieldSet(Generic[_T], FieldSet, metaclass=ABCMeta):
    """FieldSet for DecryptRequest.

    Union members may list any fields required to fulfill the
    instantiation of the `DecryptResponse` result of the decrypt
    rule.

    """

    # Subclasses must provide this, to a union member (subclass) of `DecryptRequest`.
    decrypt_request_type: ClassVar[Type[_T]]  # type: ignore[misc]

    def _request(self) -> _T:
        """Internal helper for the core decrypt goal."""
        return self.decrypt_request_type(field_set=self)

    @classmethod
    def rules(cls) -> tuple[UnionRule, ...]:
        """Helper method for registering the union members."""
        return (
            UnionRule(DecryptFieldSet, cls),
            UnionRule(DecryptRequest, cls.decrypt_request_type),
        )

    def get_output_data(self) -> DecryptOutputData:
        return DecryptOutputData({"target": self.address})


@goal_rule
async def decrypt(
    console: Console,
    decrypt_subsystem: DecryptSubsystem,
    targets: Targets,
    unions: UnionMembership,
) -> Decrypt:
    target_roots_to_decrypt_field_sets = await Get(
        TargetRootsToFieldSets,
        TargetRootsToFieldSetsRequest(
            DecryptFieldSet,
            goal_description="Generate field sets",
            no_applicable_targets_behavior=NoApplicableTargetsBehavior.ignore,
        ),
    )

    gets = []
    for tgt in targets:
        if tgt not in target_roots_to_decrypt_field_sets.mapping:
            continue

        mapping = target_roots_to_decrypt_field_sets.mapping[tgt]

        get = Get(
            DecryptResponse,
            DecryptRequest,
            mapping[0]._request(),
        )

        gets.append(get)

    responses = await MultiGet(*gets)
    with decrypt_subsystem.output(console) as write_stdout:
        for response in responses:
            secret = response.secret
            write_stdout(f"Secret {secret.name} from {secret.source}: {secret.value}\n")

    return Decrypt(exit_code=0)


def rules():
    return collect_rules()



from dataclasses import dataclass

from pants.backend.python.goals.publish import (
    PublishPythonPackageRequest,
    SkipTwineUploadField,
    twine_upload_args,
)
from pants.backend.python.subsystems.twine import TwineSubsystem
from pants.backend.python.target_types import (
    BuildBackendEnvVarsField,
    GenerateSetupField,
    InterpreterConstraintsField,
    LongDescriptionPathField,
    PythonDistributionDependenciesField,
    PythonDistributionEntryPointsField,
    PythonProvidesField,
    SDistConfigSettingsField,
    SDistField,
    WheelConfigSettingsField,
    WheelField,
)
from pants.backend.python.util_rules.pex import PexRequest, VenvPex, VenvPexProcess
from pants.core.goals.publish import (
    PublishFieldSet,
    PublishOutputData,
    PublishPackages,
    PublishProcesses,
)
from pants.core.util_rules.config_files import ConfigFiles, ConfigFilesRequest
from pants.engine.addresses import Addresses, UnparsedAddressInputs
from pants.engine.fs import CreateDigest, Digest, MergeDigests, Snapshot
from pants.engine.process import InteractiveProcess, Process
from pants.engine.rules import Get, MultiGet, collect_rules, rule
from pants.engine.target import COMMON_TARGET_FIELDS, Target, WrappedTarget, WrappedTargetRequest
from pants.option.global_options import GlobalOptions
from pants.util.docutil import doc_url
from pants.util.strutil import softwrap

from pants_backend_secrets.exception import FailedDecryptException, NoDecrypterException
from pants_backend_secrets.secret import SecretsField
from pants_backend_secrets.secret_request import (
    FallibleSecretsRequest,
    FallibleSecretsResponse,
    SecretsRequestRequest,
    SecretsRequestWrap,
)


class PublishSecretsField(SecretsField):
    alias = "repo_secrets"

    help = softwrap("Dictionary with all secrets to request. Each key should match a repository name.")


class PublishPythonWithSecretPackageRequest(PublishPythonPackageRequest):
    pass


@dataclass(frozen=True)
class PublishPythonWithSecretPackageFieldSet(PublishFieldSet):
    publish_request_type = PublishPythonWithSecretPackageRequest
    required_fields = (PublishSecretsField,)

    repositories: PublishSecretsField
    skip_twine: SkipTwineUploadField

    def get_output_data(self) -> PublishOutputData:
        return PublishOutputData(
            {
                "publisher": "twine",
                **super().get_output_data(),
            }
        )


class PythonDistributionWithSecret(Target):
    alias = "python_distribution_with_secret"
    core_fields = (
        *COMMON_TARGET_FIELDS,
        InterpreterConstraintsField,
        PythonDistributionDependenciesField,
        PythonDistributionEntryPointsField,
        PythonProvidesField,
        GenerateSetupField,
        WheelField,
        SDistField,
        WheelConfigSettingsField,
        SDistConfigSettingsField,
        BuildBackendEnvVarsField,
        LongDescriptionPathField,
        PublishSecretsField,
    )
    help = softwrap(f"""
        A publishable Python setuptools distribution (e.g. an sdist or wheel).
        See {doc_url('python-distributions')}.
        """)


@rule
async def twine_upload_with_secret(
    request: PublishPythonWithSecretPackageRequest,
    twine_subsystem: TwineSubsystem,
    global_options: GlobalOptions,
) -> PublishProcesses:
    dists = tuple(
        artifact.relpath for pkg in request.packages for artifact in pkg.artifacts if artifact.relpath
    )

    if twine_subsystem.skip or not dists:
        return PublishProcesses()

    # Too verbose to provide feedback as to why some packages were skipped?
    skip = None
    if request.field_set.skip_twine.value:
        skip = f"(by `{request.field_set.skip_twine.alias}` on {request.field_set.address})"
    elif not request.field_set.repositories.value:
        # I'd rather have used the opt_out mechanism on the field set, but that gives no hint as to
        # why the target was not applicable..
        skip = f"(no `{request.field_set.repositories.alias}` specified for {request.field_set.address})"

    if skip:
        return PublishProcesses(
            [
                PublishPackages(
                    names=dists,
                    description=skip,
                ),
            ]
        )

    twine_pex, packages_digest, config_files = await MultiGet(
        Get(VenvPex, PexRequest, twine_subsystem.to_pex_request()),
        Get(Digest, MergeDigests(pkg.digest for pkg in request.packages)),
        Get(ConfigFiles, ConfigFilesRequest, twine_subsystem.config_request()),
    )

    ca_cert_request = twine_subsystem.ca_certs_digest_request(global_options.ca_certs_path)
    ca_cert = await Get(Snapshot, CreateDigest, ca_cert_request) if ca_cert_request else None
    ca_cert_digest = (ca_cert.digest,) if ca_cert else ()

    input_digest = await Get(
        Digest, MergeDigests((packages_digest, config_files.snapshot.digest, *ca_cert_digest))
    )
    pex_proc_requests = []
    secret_requests = []

    repositories = []
    for repo, secret in request.field_set.repositories.value.items():
        repositories.append(repo)
        secret_address = await Get(
            Addresses,
            UnparsedAddressInputs(
                [secret],
                owning_address=request.field_set.address,
                description_of_origin=f"the `{secret}` from the target {request.field_set.repositories}",
            ),
        )
        wrapped_target = await Get(
            WrappedTarget,
            WrappedTargetRequest(secret_address[0], description_of_origin="twine_upload_with_secret"),
        )

        secret_request = await Get(SecretsRequestWrap, SecretsRequestRequest(wrapped_target.target))
        if secret_request.request is None:
            raise NoDecrypterException(
                f"No valid decrypter found for secret: `{secret_address[0]}` of "
                f"type `{wrapped_target.target.alias}`"
            )

        secret_requests.append(Get(FallibleSecretsResponse, FallibleSecretsRequest, secret_request.request))

    fallible_secrets = await MultiGet(*secret_requests)
    secrets = []
    for repo, maybe_secret in zip(repositories, fallible_secrets):
        if maybe_secret.exit_code != 0:
            raise FailedDecryptException(
                f"Failed decrypting secret for repo: {repo}",
                maybe_secret.stdout,
                maybe_secret.stderr,
            )

        assert maybe_secret.response is not None, "cannot be None if exit_code is 0"
        secrets.append(maybe_secret.response.value)

    for repo, secret in zip(repositories, secrets):
        pex_proc_requests.append(
            VenvPexProcess(
                twine_pex,
                argv=twine_upload_args(twine_subsystem, config_files, repo, dists, ca_cert),
                input_digest=input_digest,
                extra_env={
                    "TWINE_USERNAME": "__token__",
                    "TWINE_PASSWORD": secret.value,
                },
                description=repo,
            )
        )

    processes = await MultiGet(Get(Process, VenvPexProcess, request) for request in pex_proc_requests)

    return PublishProcesses(
        PublishPackages(
            names=dists,
            process=InteractiveProcess.from_process(process),
            description=process.description,
            data=PublishOutputData({"repository": process.description}),
        )
        for process in processes
    )


def rules():
    return [
        *collect_rules(),
        *PublishPythonWithSecretPackageFieldSet.rules(),
    ]



"""

Pants backend for handling secrets.
"""

from pants_backend_secrets import (
    env_secret_provider,
    publish_secret_python,
    secret_request,
    targets,
)
from pants_backend_secrets.goals import decrypt


def rules():
    return [
        *decrypt.rules(),
        *publish_secret_python.rules(),
        *secret_request.rules(),
        *env_secret_provider.rules(),
    ]


def target_types():
    return [*targets.targets(), publish_secret_python.PythonDistributionWithSecret]



""" """

from pants.engine.target import COMMON_TARGET_FIELDS, StringField, Target


class EnvironmentSecretKey(StringField):
    alias = "key"

    help = "The name/key for a secret in the host environment"


class EnvironmentSecret(Target):
    alias = "env_secret"
    core_fields = (
        *COMMON_TARGET_FIELDS,
        EnvironmentSecretKey,
    )

    help = "A declaration for an environment key that can be read as a secret."


def targets():
    return [
        EnvironmentSecret,
    ]
